import logging
import time
from functools import wraps


def timed(func):
    """This decorator prints the execution time for the decorated function."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        logging.info(f"Launching {func.__name__}")
        result = func(*args, **kwargs)
        end = time.time()
        logging.info(f"{func.__name__} ran in {round(end - start, 2)}s")
        return result

    return wrapper



# coding=utf-8
"""Top-level package for Bolster."""

__author__ = """Andrew Bolster"""
__email__ = "me@andrewbolster.info"
__version__ = "0.1.2"

import base64
import contextlib
import gzip
import json
import logging
import os
import random
import sys
import time
import traceback
from collections import Counter, defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed, Future
from functools import wraps, partial
from itertools import chain, islice, groupby
from operator import itemgetter
from pathlib import Path
from typing import (
    Sequence,
    Generator,
    Iterable,
    List,
    Dict,
    Iterator,
    Union,
    AnyStr,
    Optional,
    Callable,
    SupportsInt,
    SupportsFloat,
    Tuple,
    Set,
    Hashable,
    Any,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def _dumb_passthrough(x, **kwargs):
    """Pointless passthrough replacement for tqdm (and similar) fallback

    Args:
      x: return:

    Returns:

    """
    return x


def always(x, **kwargs) -> bool:
    """Pointless passthrough replacement for 'always true' filtering

    >>> always('false')
    True
    >>> always(False)
    True
    >>> always(True)
    True
    """

    return True


def poolmap(
    f: Callable,
    iterable: Iterable,
    max_workers: Optional[int] = None,
    progress: Callable = None,
    **kwargs,
) -> Dict:
    """Helper function to encapsulate a ThreadPoolExecutor mapped function workflow
    Accepts (assumed to be `tqdm` style) progress monitor callback

    `kwargs` are passed identically to all `f(i)` calls for each i in `iterable`

    Args:
      f: function to map across
      iterable:
      max_workers: (Default value = None)
      progress: (Default value = None)
      **kwargs: passed as arguments to f

    Returns:

    """
    futures = {}
    results = {}

    if progress is None:
        progress = _dumb_passthrough

    with ThreadPoolExecutor(max_workers=max_workers) as exc:
        for arg in iterable:
            if arg not in results:
                futures[exc.submit(f, arg, **kwargs)] = arg
        for future in progress(as_completed(futures), total=len(futures)):
            arg = futures[future]
            results[arg] = future.result()

    return results


def batch(seq: Sequence, n: int = 1) -> Generator[Iterable, None, None]:
    """Split a sequence into n-length batches (is still iterable, not list)

    Args:
      seq:
      n:  (Default value = 1)

    Returns:

    >>> next((b for b in batch(range(10), 2)))
    range(0, 2)
    >>> [b for b in batch(list(range(10)), 2)]
    [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
    """
    parent_length = len(seq)
    for ndx in range(0, parent_length, n):
        yield seq[ndx : min(ndx + n, parent_length)]


def chunks(iterable: Iterable, size=10) -> Generator[List, None, None]:
    """Outputs <list> chunks of size N from an iterable (generator)

    Args:
      iterable: param size:
      iterable: Iterable:
      size:  (Default value = 10)

    Returns:

    >>> next((b for b in chunks(range(10), 2)))
    [0, 1]
    >>> [b for b in chunks(list(range(10)), 2)]
    [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
    """

    iterator = iter(iterable)
    for first in iterator:
        yield list(chain([first], islice(iterator, size - 1)))


def arg_exception_logger(func: Callable) -> Callable:
    """Helper Decorator to provide info on the arguments that cause the exception of a wrapped function

    Args:
      func:

    Returns:

    """

    # noinspection PyMissingOrEmptyDocstring
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            raise ValueError(f"Failed with args {args} and kwargs {kwargs}") from e

    return wrapper


# noinspection PyShadowingNames
def backoff(
    exception_to_check: Union[BaseException, Sequence[BaseException]],
    tries: SupportsInt = 5,
    delay: SupportsFloat = 0.2,
    backoff: SupportsFloat = 2,
    logger: Optional[logging.Logger] = logger,
):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    Args:
      exception_to_check: the exception to check. may be a tuple of
    exceptions to check
      tries: number of times to try (not retry) before giving up (Default value = 5)
      delay: initial delay between retries in seconds (Default value = 0.4)
      backoff: backoff multiplier e.g. value of 2 will double the delay
    each retry (Default value = 2)
      logger: logger to use. If None, print (Default value = local utils logger)

    Returns:

    """

    # noinspection PyMissingOrEmptyDocstring
    def deco_retry(f):
        # noinspection PyMissingOrEmptyDocstring
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            mdelay += (random.random() - 0.5) * (delay / 2)
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except exception_to_check as e:
                    msg = f"{e}, Retrying in {mdelay} seconds..."
                    if logger:
                        logger.warning(msg)
                    else:
                        print(msg)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry


class MultipleErrors(BaseException):
    """Exception Class to enable the capturing of multiple exceptions without interrupting control flow, i.e. catch
    the exception, but carry on and report the exceptions at the end.

    E.g.

    .. code-block:: python

        exceptions = MultipleErrors()
        try:
            do_risky_thing_with(this) #raises ValueError
        except:
            exceptions.capture_current_exception()
        try:
            do_other_thing_with(this) #raises AttributeError
        except:
            exceptions.capture_current_exception()
        exceptions.do_raise()



    .. code-block:: none

         Traceback (most recent call last):
            ....
        Value Error

        Traceback (most recent call last):
            ...
        AttributeError

    """

    def __init__(self, errors=None):
        self.errors = errors or []

    @classmethod
    def _traceback_for(cls, exc_info):
        """Formatting!"""
        return "".join(traceback.format_exception(*exc_info))

    def __str__(self):
        tracebacks = "\n\n".join(
            self._traceback_for(exc_info) for exc_info in self.errors
        )
        parts = ("See the following exception tracebacks:", "=" * 78, tracebacks)
        msg = "\n".join(parts)
        return msg

    def capture_current_exception(self):
        """Gathers exception info from the current context and retains it"""
        self.errors.append(sys.exc_info())

    def do_raise(self):
        """Raises itself if it contains any errors"""
        if self.errors:
            raise self


def tag_gen(seq: Iterator[Dict], **kwargs) -> Iterator[Dict]:
    """Generator stream that adds a kwargs to each entry yielded

    The below example shows the creation of an empty dict generator where
    tag_gen is used to insert a new key/value (k=1) in each item on the fly

    >>> all([i['k'] == 1 for i in tag_gen(({} for _ in range(4)), k=1)])
    True

    Args:
      seq: param kwargs:
      seq: Iterator[Dict]:
      **kwargs:

    """
    for item in seq:
        new_item = item
        for k, v in kwargs.items():
            new_item[k] = v
        yield new_item


def exceptional_executor(
    futures: Sequence[Future], exception_handler=None, timeout=None
) -> Iterator:
    """Generator for concurrent.Futures handling

    When an exception is raised in an executing Future, f.result() called on it's own will raise that
    exception in the parent thread, killing execution and causing loss of 'future local' scope.

    Instead, query the future for it's exception state first, and handle that separately, by default
    by logging it as an exception.


    Args:
        futures:
        exception_handler:
        timeout:

    Returns:

    """

    def default_hdl(e: BaseException, f: Future):
        try:
            raise e
        except BaseException:
            logging.exception(f"Caught exception in a future: {f}")

    if exception_handler is None:
        exception_handler = default_hdl

    for f in as_completed(futures, timeout=timeout):
        e = f.exception()
        if e is not None:
            exception_handler(e, f)
        else:
            yield f.result()


@contextlib.contextmanager
def working_directory(path: Union[str, Path]) -> Generator:
    """Contextmanager that changes working directory and returns to previous on exit.

    Args:
      path: Union[str: Path]:

    """
    prev_cwd = Path.cwd()
    os.chdir(str(path))
    try:
        yield
    finally:
        os.chdir(str(prev_cwd.absolute()))


def compress_for_relay(obj: Union[List, Dict]) -> AnyStr:
    """Compress json-serializable object to a gzipped base64 string

    Args:
      obj: return:
      obj: Union[List,Dict]:

    >>> decompress_from_relay(compress_for_relay(['test']))
    ['test']

    >>> decompress_from_relay(compress_for_relay({'test':'test'}))
    {'test': 'test'}

    """
    return base64.b64encode(gzip.compress(json.dumps(obj).encode("utf-8"), 6)).decode()


def decompress_from_relay(msg: AnyStr) -> Union[List, Dict]:
    """Uncompress  gzipped base64 string to a json-serializable object
    ['test']

    Args:
      msg: AnyStr:

    Returns:

    """
    return json.loads(gzip.decompress(base64.b64decode(msg)))


class memoize(object):
    """cache the return value of a method

    This class is meant to be used as a decorator of methods. The return value
    from a given method invocation will be cached on the instance whose method
    was invoked. All arguments passed to a method decorated with memoize must
    be hashable.

    If a memoized method is invoked directly on its class the result will not
    be cached. Instead the method will be invoked like a static method:

    .. code-block:: python

        class Obj(object):
            @memoize
            def add_to(self, arg):
            return self + arg

        Obj.add_to(1) # not enough arguments
        Obj.add_to(1, 2) # returns 3, result is not cached

    Source: http://code.activestate.com/recipes/577452-a-memoize-decorator-for-instance-methods/

    Augmented with cache hit/miss population Counters

    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
            obj.__hits = Counter()
            obj.__misses = Counter()
        key = (self.func, args[1:], frozenset(kw.items()))
        try:
            res = cache[key]
            obj.__hits[key] += 1
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
            obj.__misses[key] += 1
        return res


def pretty_print_request(
    req, expose_auth=False, authentication_header_blacklist: Optional[Sequence] = None
) -> None:
    """At this point it is completely built and ready
    to be fired; it is "prepared".

    However pay attention at the formatting used in
    this function because it is programmed to be pretty
    printed and may differ from the actual request.

    Args:
      req:
      expose_auth:  (Default value = False)

    Returns:

    """
    printable_headers = {k: v for k, v in req.headers.items()}
    if not expose_auth:
        for header in authentication_header_blacklist:
            if header in printable_headers.keys():
                printable_headers[header] = "<<REDACTED>>"
    print(
        f"-----------START-----------\n" f"{req.method} {req.url}\n",
        "\n".join("{}: {}".format(k, v) for k, v in printable_headers.items()),
    )
    if req.body is not None:
        print(req.body)


def get_recursively(search_dict: Dict, field: str) -> List:
    """Takes a dict with nested lists and dicts,
    and searches all dicts for a key of the field
    provided.

    Originally taken from https://stackoverflow.com/a/20254842

    Args:
      search_dict: Dict:
      field: str:

    Returns:

    >>> get_recursively({'id' : 5,'children' : {'id' : 6,'children' : {'id' : 7,'children' : {}}}}, 'id')
    [5, 6, 7]
    """
    fields_found = []

    for key, value in search_dict.items():

        if key == field:
            fields_found.append(value)

        elif isinstance(value, dict):
            results = get_recursively(value, field)
            for result in results:
                fields_found.append(result)

        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    more_results = get_recursively(item, field)
                    for another_result in more_results:
                        fields_found.append(another_result)

    return fields_found


def transform_(r: Dict, rule_keys: Dict[AnyStr, Optional[Tuple]]) -> Dict:
    """
    Generic Item-wise transformation function;
    The values in `r` are updated based on key-matching in `rule_keys`,
    i.e. -> out[k] = rule_keys[k] (r[k])

    HOWEVER, this can do more that straight callable mapping; can *also* update
    the key, i.e., for a given rule such that `R = rule_keys[k]`:

    R can be used to select that field to be selected in the output
    >>> r = {'a':'1','b':'2','c':'3'}
    >>> transform_(r, {'a':None})
    {'a': '1'}

    Rename a key
    >>> transform_(r, {'a':('A',None)})
    {'A': '1'}

    Apply a function to a key's value
    >>> transform_(r, {'a':('a',int)})
    {'a': 1}

    Or a combination of these
    >>> transform_(r, {'a':('A',int), 'b':None})
    {'A': 1, 'b': '2'}

    """
    out_record = {}
    for k, v in rule_keys.items():
        f = lambda x: x  # noqa: E731

        if v is None:
            new_k = k
        else:
            try:
                new_k, f = v
            except ValueError:
                logger.exception("Rules must have either None value or a 2-tuple")
                raise

            if new_k is None:
                new_k = k
            if f is None:
                f = lambda x: x  # noqa: E731

        out_record[new_k] = f(r[k]) if k in r else None

    return out_record


def diff(new: Dict, old: Dict, excluded_fields: Optional[set] = None) -> Dict:
    """
    Perform a one-depth diff of a pair of dictionaries

    #TODO diff needs tests
    """
    if excluded_fields is None:
        excluded_fields = set()

    diffed_values = {
        k: {"old": old.get(k, None), "new": new.get(k, None)}
        for k in set(new.keys()).union(old.keys()) - excluded_fields
        if old.get(k, None) != new.get(k, None)
    }
    return diffed_values


def aggregate(
    base: List[Dict],
    group_key: Union[AnyStr, Tuple[AnyStr], List[AnyStr]],
    item_key: AnyStr,
    condition: Optional[Callable] = None,
):
    """
    Abstracted groupby-sum for lists of dicts
    operationally equivalent to
    # TODO aggregate needs tests
    ```
    df = pd.DataFrame(base)
    df.where(condition).groupby(group_key)[item_key].sum()
    ```

    Args:
        base:
        group_key:
        item_key:
        condition:

    Returns:

    """
    agg_c = Counter()
    if condition is None:
        condition = lambda x: True  # noqa: E731

    if isinstance(group_key, (tuple, list)):
        grouper = itemgetter(*group_key)
    else:
        grouper = itemgetter(group_key)

    for source_key, g in groupby(filter(condition, base), grouper):
        for sig in g:
            agg_c[source_key] += sig[item_key]
    agg_d = dict(sorted(agg_c.items(), key=itemgetter(1), reverse=True))
    return agg_d


def breadth(d):
    """
    Get the total 'width' of a tree

    > Why was this a thing? No idea

    """
    if isinstance(d, dict):
        width = sum(map(breadth, d.values()))
    else:
        width = 1
    return width


def depth(d: Dict[Any, int]) -> int:
    """
    Get the maximum depth of a tree
    """
    if isinstance(d, dict):
        height = max(map(depth, d.values())) + 1
    else:
        height = 0
    return height


def set_keys(d: Dict) -> Set:
    """
    Extract the set of all keys of a nested dict/tree
    """
    keys = set()
    for k, v in d.items():
        if isinstance(v, dict):
            keys.update(set_keys(v))
        else:
            keys.update([k])
    return keys


def keys_at(d: Dict, n: SupportsInt, i: SupportsInt = 0) -> Iterator:
    """
    Extract the keys of a tree at a given depth
    """
    if isinstance(d, dict):
        for k, v in d.items():
            if i == n:
                yield k
            else:
                yield from keys_at(v, n, i + 1)


def items_at(d: Dict, n: SupportsInt, i: SupportsInt = 0) -> Iterator[Tuple]:
    """
    Extract the elements from a tree at a given depth
    """
    if isinstance(d, dict):
        for k, v in d.items():
            if i == n:
                yield k, v
            else:
                yield from items_at(v, n, i + 1)


def leaves(d: Dict) -> Iterator:
    """
    Iterate on the leaves of a tree
    """
    if isinstance(d, dict):
        for k, v in d.items():
            yield from leaves(v)
    else:
        yield (d)


def leaf_paths(d: Dict, path: Optional[List] = None) -> Iterator[Tuple[List, Dict]]:
    if path is None:
        path = []
    if isinstance(d, dict):
        for k, v in d.items():
            yield from leaf_paths(v, path + [k])
    else:
        yield (path, d)


def flatten_dict(d: Dict, head: str = "", sep: str = ":") -> Dict:
    new_d = {}
    for k, v in d.items():
        if isinstance(v, dict):
            new_d.update(
                flatten_dict(
                    # Sort the inner items first to have alpha order but preserve outer structure
                    dict(sorted(v.items())),
                    head=k,
                )
            )
        else:
            new_d[sep.join([head, k])] = v
    return new_d


def uncollect_object(d: Dict) -> Dict:
    new_d = {}
    for k, v in d.items():
        if isinstance(v, (defaultdict, Counter)):
            new_d[k] = uncollect_object(v)
        else:
            new_d[k] = v
    return new_d


def dict_concat_safe(
    d: Dict, keys: List[Hashable], default: Optional[Any] = None
) -> Iterator:
    """
    Really Lazy Func because `dict.get('key',default)` is a pain in the ass for lists
    """
    for k in keys:
        yield d.get(k, default)


def build_default_mapping_dict_from_keys(keys: List[str]) -> Dict[str, str]:
    """
    Constructs a mapping dictionary between (presumably) snakecase keys to 'human-readable' title case

    Intended for easy construction of presentable graphs/tables etc.

    >>> build_default_mapping_dict_from_keys(['a_b','b_c','c_d'])
    {'a_b': 'A B', 'b_c': 'B C', 'c_d': 'C D'}
    """
    return dict([(f, f.replace("_", " ").title()) for f in keys])



from typing import AnyStr

import pandas as pd


def add_totals(
    df: pd.DataFrame,
    column_total: AnyStr = "total",
    row_total: AnyStr = "total",
    inplace=True,
):
    """
    Add Row and Column totals to a dataframe (in place)

    >>> add_totals(pd.DataFrame([[0,1,2],[3,4,5]]))
           0  1  2  total
    0      0  1  2      3
    1      3  4  5     12
    total  3  5  7     15

    >>> add_totals(pd.DataFrame([[0,1,2],[3,4,5]]),'ctot', 'rtot')
          0  1  2  rtot
    0     0  1  2     3
    1     3  4  5    12
    ctot  3  5  7    15


    >>> df = pd.DataFrame([[0,1,2],[3,4,5]])
    >>> add_totals(df, inplace=False)
           0  1  2  total
    0      0  1  2      3
    1      3  4  5     12
    total  3  5  7     15

    >>> df
       0  1  2
    0  0  1  2
    1  3  4  5

    """
    if not inplace:
        df = df.copy(deep=True)
    df.loc[column_total] = df.sum(numeric_only=True, axis=0)
    df.loc[:, row_total] = df.sum(numeric_only=True, axis=1)

    return df


def fix_datetime_tz_columns(df: pd.DataFrame, inplace=True) -> pd.DataFrame:
    """
    Strip Timezone information from relevant datetime columns in a dataframe

    Parameters
    ----------
    df
    inplace (bool)


    Returns
    -------
    df

    """
    if not inplace:
        df = df.copy(deep=True)
    date_columns = df.select_dtypes(
        include=["datetime64[ns, UTC]", "datetimetz"]
    ).columns
    for date_column in date_columns:
        df[date_column] = df[date_column].dt.tz_localize(None)
    return df



import logging

import tqdm


class TqdmLoggingHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)



import csv
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import Text

import bs4
import requests
from tqdm.auto import tqdm

from .. import always
from .. import dict_concat_safe
from ..utils.web import download_extract_zip


def get_basic_company_data_url() -> Text:
    """
    Parse the companies house website to get the current URL for the 'BasicCompanyData'

    Currently uses the 'one file' method but it could be split into the multi files for memory efficiency
    """
    base_url = "http://download.companieshouse.gov.uk/en_output.html"
    s = bs4.BeautifulSoup(requests.get(base_url).content)
    for a in s.find_all("a"):
        if a.get("href").startswith("BasicCompanyDataAsOneFile"):
            url = f"http://download.companieshouse.gov.uk/{a.get('href')}"
            break  # assume first time lucky

    return url


def query_basic_company_data(
    query_func: Callable[..., bool] = always
) -> Iterator[Dict]:
    """
    Grab the url for the basic company data, and walk through the CSV files within, and
    for each row in each CSV file, parse the row data through the given `query_func`
    such that if `query_func(row)` is True it will be yielded
    """
    url = get_basic_company_data_url()
    for filename, data in tqdm(download_extract_zip(url)):
        for row in tqdm(csv.DictReader((d.decode("utf-8") for d in data))):
            if query_func(row):
                yield row


def companies_house_record_might_be_farset(r: Dict) -> bool:
    """
    A heuristic function for working out if a record in the companies house registry *might* be based in Farset Labs
    Almost certainly incomplete and needs more testing/validation
    """
    if r["RegAddress.PostCode"].lower().replace(" ", "") != "bt125gh":
        return False
    address_line = ",".join(
        map(
            str,
            dict_concat_safe(
                r,
                [
                    "RegAddress.CareOf",
                    "RegAddress.AddressLine1",
                    "RegAddress.AddressLine2",  # This appears to be optional now
                ],
                default="",
            ),
        )
    ).lower()
    if "farset" in address_line:
        return True
    elif "unit 10" in address_line:
        return False
    elif "unit 18" in address_line:
        return False
    elif "unit 17" in address_line:
        return False
    elif "unit 1" in address_line:
        return True
    else:
        return False


def get_companies_house_records_that_might_be_in_farset() -> Iterator[Dict]:
    yield from query_basic_company_data(companies_house_record_might_be_farset)



import warnings
from time import time

import numpy as np
import pandas as pd
import scipy.stats as st
from tqdm.auto import tqdm


# Create models from data https://stackoverflow.com/questions/6620471/fitting-empirical-distribution-to-theoretical-ones-with-scipy-python
def best_fit_distribution(
    data, bins=200, ax=None, include_slow=False, discriminator="sse"
):
    """Model data by finding best fit distribution to data"""
    # Get histogram of original data
    y, x = np.histogram(data, bins=bins, density=True)
    x = (x + np.roll(x, -1))[:-1] / 2.0

    # Distributions to check
    DISTRIBUTIONS = [
        st.alpha,
        st.anglit,
        st.arcsine,
        st.beta,
        st.betaprime,
        st.bradford,
        st.burr,
        st.cauchy,
        st.chi,
        st.chi2,
        st.cosine,
        st.dgamma,
        st.dweibull,
        st.erlang,
        st.expon,
        st.exponnorm,
        st.exponweib,
        st.exponpow,
        st.f,
        st.fatiguelife,
        st.fisk,
        st.foldcauchy,
        st.foldnorm,
        st.frechet_r,
        st.frechet_l,
        st.genlogistic,
        st.genpareto,
        st.gennorm,
        st.genexpon,
        st.genextreme,
        st.gausshyper,
        st.gamma,
        st.gengamma,
        st.genhalflogistic,
        st.gilbrat,
        st.gompertz,
        st.gumbel_r,
        st.gumbel_l,
        st.halfcauchy,
        st.halflogistic,
        st.halfnorm,
        st.halfgennorm,
        st.hypsecant,
        st.invgamma,
        st.invgauss,
        st.invweibull,
        st.johnsonsb,
        st.johnsonsu,
        st.ksone,
        st.kstwobign,
        st.laplace,
        st.levy,
        st.levy_l,
        st.logistic,
        st.loggamma,
        st.loglaplace,
        st.lognorm,
        st.lomax,
        st.maxwell,
        st.mielke,
        st.nakagami,
        st.ncx2,
        st.ncf,
        st.nct,
        st.norm,
        st.pareto,
        st.pearson3,
        st.powerlaw,
        st.powerlognorm,
        st.powernorm,
        st.rdist,
        st.reciprocal,
        st.rayleigh,
        st.rice,
        st.recipinvgauss,
        st.semicircular,
        st.t,
        st.triang,
        st.truncexpon,
        st.truncnorm,
        st.tukeylambda,
        st.uniform,
        st.vonmises,
        st.vonmises_line,
        st.wald,
        st.weibull_min,
        st.weibull_max,
        st.wrapcauchy,
    ]
    SLOW_DISTRIBUTIONS = [st.levy_stable]

    if include_slow:
        DISTRIBUTIONS += SLOW_DISTRIBUTIONS

    # Best holders
    best_distribution = st.norm
    best_params = (0.0, 1.0)
    best_discriminator_value = np.inf

    times = {}

    # Estimate distribution parameters from data
    for distribution in tqdm(DISTRIBUTIONS):

        # Try to fit the distribution
        try:
            # Ignore warnings from data that can't be fit
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore")

                # fit dist to data
                start = time()
                print(distribution.name, end=" ")
                params = distribution.fit(data)
                times[distribution.name] = time() - start
                print(f"took {int(times[distribution.name])}")

                # Separate parts of parameters
                arg = params[:-2]
                loc = params[-2]
                scale = params[-1]

                if discriminator == "sse":
                    # Calculate fitted PDF and error with fit in distribution
                    pdf = distribution.pdf(x, loc=loc, scale=scale, *arg)
                    discriminator_value = np.sum(np.power(y - pdf, 2.0))
                else:
                    raise RuntimeError(
                        "You didn't finish this and you were planning on doing KS discrimination next"
                    )

                # if axis pass in add to plot
                try:
                    if ax:
                        pd.Series(pdf, x).plot(ax=ax)
                except Exception:
                    pass

                # identify if this distribution is better
                if best_discriminator_value > discriminator_value > 0:
                    best_distribution = distribution
                    best_params = params
                    best_discriminator_value = discriminator_value
                    print(
                        f"New best, {distribution.name} and got an sse of {discriminator_value}"
                    )

        except Exception:
            pass

    return (best_distribution.name, best_params)



"""
Working with the Norther Ireland House Price index data

Original Source: https://www.nisra.gov.uk/statistics/housing-community-and-regeneration/northern-ireland-house-price-index

See [here](https://andrewbolster.info/2022/03/NI-House-Price-Index.html) for more details

Generic problems fixed;

* Tables offset from header with notes and annotations
* Inconsistent offset
* Inconsistent treatment of quarterly periods (Q1 2020/Quarter 1 2020/ (2020,Q1) etc)
* Inconsistent header alignment (i.e. leaving sub-category annotations on what is actually the period index header)


"""
import re
from typing import Dict
from typing import Text

import bs4
import pandas as pd
import requests

DEFAULT_URL = "https://www.finance-ni.gov.uk/publications/ni-house-price-index-statistical-reports"
TABLE_TRANSFORMATION_MAP = {}


def pull_source(base_url=DEFAULT_URL) -> Dict[Text, pd.DataFrame]:
    """
    Pull raw NI House Price Index Excel from finance-ni.gov.uk listing

    Parameters
    ----------
    base_url

    Returns
    -------

    """
    base_content = requests.get(base_url).content
    base_soup = bs4.BeautifulSoup(base_content)
    source_url = None
    for a in base_soup.find_all("a"):
        if a.attrs.get("href", "").endswith("xlsx"):
            source_url = a.attrs["href"]

    if source_url is not None:
        source_df = pd.read_excel(source_url, sheet_name=None)  # Load all worksheets in
    else:
        raise RuntimeError(
            f"Could not find valid/relevant Excel source file on {base_url}"
        )

    return source_df


def basic_cleanup(df: pd.DataFrame, offset=1) -> pd.DataFrame:
    """
    Generic cleanup operations for NI HPI data;
    * Re-header from Offset row and translate table to eliminate incorrect headers
    * remove any columns with 'Nan' or 'None' in the given offset-row
    * If 'NI' appears and all the values are 100, remove it.
    * Remove any rows below and including the first 'all nan' row (gets most tail-notes)
    * If 'Sale Year','Sale Quarter' appear in the columns, replace with 'Year','Quarter' respectively
    * For Year; forward fill any none/nan values
    * If Year/Quarter appear, add  a new composite 'Period' column with a PeriodIndex columns representing the
        year/quarter (i.e. 2022-Q1)
    * Reset and drop the index
    * Attempt to infer the new/current column object types

    Parameters
    ----------
    df
    offset

    Returns
    -------

    """
    df = df.copy()
    # Re-header from row 1 (which was row 3 in excel)
    new_header = df.iloc[offset]
    df = df.iloc[offset + 1 :]
    df.columns = new_header

    # remove 'NaN' trailing columns
    df = df[df.columns[pd.notna(df.columns)]]

    # 'NI' is a usually hidden column that appears to be a checksum;
    # if it's all there and all 100, remove it, otherwise, complain.
    # (Note, need to change this 'if' logic to just 'if there's a
    # column with all 100's, but cross that bridge later)
    if "NI" in df:
        assert (
            df["NI"].all() and df["NI"].mean() == 100
        ), "Not all values in df['NI'] == 100"
        df = df.drop("NI", axis=1)

    # Strip rows below the first all-nan row, if there is one
    # (Otherwise this truncates the tables as there is no
    # idxmax in the table of all 'false's)
    if any(df.isna().all(axis=1)):
        idx_first_bad_row = df.isna().all(axis=1).idxmax()
        df = df.loc[: idx_first_bad_row - 1]

    # By Inspection, other tables use 'Sale Year' and 'Sale Quarter'
    if set(df.keys()).issuperset({"Sale Year", "Sale Quarter"}):
        df = df.rename(columns={"Sale Year": "Year", "Sale Quarter": "Quarter"})

    # For 'Year','Quarter' indexed pages, there is an implied Year
    # in Q2/4, so fill it downwards
    if set(df.keys()).issuperset({"Year", "Quarter"}):
        df["Year"] = df["Year"].astype(float).fillna(method="ffill").astype(int)

        # In Pandas we can represent Y/Q combinations as proper datetimes
        # https://stackoverflow.com/questions/53898482/clean-way-to-convert-quarterly-periods-to-datetime-in-pandas
        df.insert(
            loc=0,
            column="Period",
            value=pd.PeriodIndex(
                df.apply(lambda r: f"{r.Year}-{r.Quarter}", axis=1), freq="Q"
            ),
        )

    # reset index, try to fix dtypes, etc, (this should be the last
    # operation before returning!
    df = df.reset_index(drop=True).infer_objects()

    return df


def cleanup_contents(df: pd.DataFrame) -> pd.DataFrame:
    """
    Fix Contents table of NI HPI Stats
    * Shift/rebuild headers
    * Strip Figures because they're gonna be broken anyway

    Parameters
    ----------
    df

    Returns
    -------

    """
    new_header = df.iloc[0]
    df = df[1:].copy()
    df.columns = [*new_header[:-1], "Title"]
    # df['Worksheet Name'] = df['Worksheet Name'].str.replace('Figure', 'Fig')
    df = df[df["Worksheet Name"].str.startswith("Table")]

    return df


TABLE_TRANSFORMATION_MAP["Contents"] = cleanup_contents
# NI HPI Trends
TABLE_TRANSFORMATION_MAP["Table 1"] = basic_cleanup


def cleanup_price_by_property_type_agg(df: pd.DataFrame) -> pd.DataFrame:
    """
    NI HPI & Standardised Price Statistics by Property Type (Aggregate Table)

    Standard cleanup with a split to remove trailing index date data

    Parameters
    ----------
    df

    Returns
    -------

    """
    df = basic_cleanup(df)
    df.columns = [c.split("\n")[0] for c in df.columns]
    return df


# NI HPI & Standardised Price Statistics by Property Type
TABLE_TRANSFORMATION_MAP["Table 2"] = cleanup_price_by_property_type_agg


def cleanup_price_by_property_type(df: pd.DataFrame) -> pd.DataFrame:
    """
    NI HPI & Standardised Price Statistics by Property Type (Per Class)

    Standard cleanup, removing the property class from the table columns

    Parameters
    ----------
    df

    Returns
    -------

    """
    df = basic_cleanup(df)
    new_columns = []
    for c in df.columns:
        if c.endswith("Price Index"):
            new_columns.append("Index")
        elif c.endswith("Standardised Price"):
            new_columns.append("Price")
        else:
            new_columns.append(c)

    df.columns = new_columns

    return df


# NI {property type} Property Price Index
TABLE_TRANSFORMATION_MAP[re.compile("Table 2[a-z]")] = cleanup_price_by_property_type

# NI HPI & Standardised Price Statistics by New/Existing Resold Dwelling Type
TABLE_TRANSFORMATION_MAP["Table 3"] = cleanup_price_by_property_type_agg
TABLE_TRANSFORMATION_MAP[
    re.compile("Table 3[a-z]")
] = cleanup_price_by_property_type_agg


def cleanup_with_munged_quarters_and_total_rows(df: pd.DataFrame) -> pd.DataFrame:
    """
    Number of Verified Residential Property Sales

    * Regex 'Quarter X' to 'QX' in future 'Sales Quarter' column
    * Drop Year Total rows
    * Clear any Newlines from the future 'Sales Year' column
    * call `basic_cleanup` with offset=3

    Parameters
    ----------
    df

    Returns
    -------

    """
    df = df.copy()
    df.iloc[:, 1] = df.iloc[:, 1].str.replace("Quarter ([1-4])", r"Q\1", regex=True)
    df = df[~df.iloc[:, 1].str.contains("Total").fillna(False)]
    # Lose the year new-lines (needs astype because non str lines are
    # correctly inferred to be ints, so .str methods nan-out
    with pd.option_context("mode.chained_assignment", None):
        df.iloc[:, 0] = df.iloc[:, 0].astype(str).str.replace("\n", "")

    df = basic_cleanup(df, offset=3)
    return df


# Table 4: Number of Verified Residential Property Sales
TABLE_TRANSFORMATION_MAP["Table 4"] = cleanup_with_munged_quarters_and_total_rows


def cleanup_with_LGDs(df):
    """
    Standardised House Price & Index for each Local Government District Northern Ireland
    * Build multi-index of LGD / Metric [Index,Price]

    """
    # Basic Cleanup first
    df = basic_cleanup(df)
    # Build multi-index of LGD / Metric [Index,Price]
    # Two inner-columns per LGD
    lgds = (
        df.columns[3:]
        .str.replace(" Standardised HPI", " HPI")
        .str.replace(" HPI", "")
        .str.replace(" Standardised Price", "")
        .unique()
    )
    df.columns = [
        *df.columns[:3],
        *pd.MultiIndex.from_product(
            [lgds, ["Index", "Price"]], names=["LGD", "Metric"]
        ),
    ]
    return df


# Table 5: Standardised House Price & Index for each Local Government District Northern Ireland
TABLE_TRANSFORMATION_MAP["Table 5"] = cleanup_with_LGDs


def cleanup_merged_year_quarters_and_totals(df):
    """
    Table 5a: Number of Verified Residential Property Sales by Local Government District
    * Parse the 'Sale Year/Quarter' to two separate cols
    * Insert future-headers for Quarter and Year cols
    * Remove rows with 'total' in the first column
    * Disregard the 'Sale Year/Quarter' column
    * perform `basic_cleanup` with offset=2
    """
    # Safety first
    df = df.copy()

    # Extract 'Quarter' and 'Year' columns from the future 'Sale Year/Quarter' column
    dates = (
        df.iloc[:, 0]
        .str.extract("(Q[1-4]) ([0-9]{4})")
        .rename(columns={0: "Quarter", 1: "Year"})
    )
    for c in [
        "Quarter",
        "Year",
    ]:  # insert the dates in order, so they come out in reverse in the insert
        df.insert(1, c, dates[c])
        df.iloc[
            2, 1
        ] = c  # Need to have the right colname for when `basic_cleanup` is called.

    # Remove 'total' rows from the future 'Sale Year/Quarter' column
    df = df[~df.iloc[:, 0].str.contains("Total").fillna(False)]

    # Remove the 'Sale Year/Quarter' column all together
    df = df.iloc[:, 1:]

    # Standard cleanup
    df = basic_cleanup(df, offset=2)

    return df


# Table 5a: Number of Verified Residential Property Sales by Local Government District
TABLE_TRANSFORMATION_MAP["Table 5a"] = cleanup_merged_year_quarters_and_totals
# Table 6: Standardised House Price & Index for all Urban and Rural areas in NI
TABLE_TRANSFORMATION_MAP["Table 6"] = basic_cleanup


def cleanup_missing_year_quarter(df):
    """
    Table 7: Standardised House Price & Index for Rural Areas of Northern Ireland by drive times
    * Insert Year/Quarter future-headers
    * Clean normally
    # TODO THIS MIGHT BE VALID FOR MULTIINDEXING ON DRIVETIME/[Index/Price]
    """
    df = df.copy()
    df.iloc[1, 0] = "Year"
    df.iloc[1, 1] = "Quarter"
    df = basic_cleanup(df)
    return df


# Table 7: Standardised House Price & Index for Rural Areas of Northern Ireland by drive times
TABLE_TRANSFORMATION_MAP["Table 7"] = cleanup_missing_year_quarter

# Table 8: Number of Verified Residential Property Sales of properties in urban and rural areas and properties in rural areas by drive times witihn towns of 10,000 or more and within 1 hour of Belfast
TABLE_TRANSFORMATION_MAP["Table 8"] = cleanup_merged_year_quarters_and_totals

# Table 9: NI Average Sales Prices
TABLE_TRANSFORMATION_MAP["Table 9"] = basic_cleanup
TABLE_TRANSFORMATION_MAP[re.compile("Table 9[a-z]")] = cleanup_missing_year_quarter

# Table 10x: Number of Verified Residential Property Sales by Type in {LDG}
TABLE_TRANSFORMATION_MAP[
    re.compile("Table 10[a-z]")
] = cleanup_merged_year_quarters_and_totals


def cleanup_source(source_df: Dict[Text, pd.DataFrame]) -> Dict[Text, pd.DataFrame]:
    """
    Cleanup all the tables from the NI Housing Price Index, conforming to the best attempt at a 'standard'

    Parameters
    ----------
    source_df

    Returns
    -------

    """
    dest_df = {}
    for table_key, table_transformer in TABLE_TRANSFORMATION_MAP.items():
        if isinstance(table_key, re.Pattern):
            for table in source_df:
                if table_key.match(table):
                    dest_df[table] = table_transformer(source_df[table])
        else:
            dest_df[table_key] = table_transformer(source_df[table_key])

    return dest_df


def build():
    """
    Pulls and Cleans up the latest Northern Ireland House Price Index Data

    Returns
    -------

    """
    source_dfs = pull_source()
    dest_dfs = cleanup_source(source_dfs)
    return dest_dfs



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bolster',
 'bolster.data_sources',
 'bolster.stats',
 'bolster.utils',
 'bolster.utils.aws']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.18.29,<2.0.0',
 'bs4>=0.0.1,<0.0.2',
 'click-log>=0.3.2,<0.5.0',
 'click>=8.0.1,<9.0.0',
 'numpy>=1.21.2,<2.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pandas>=1.3.2,<2.0.0',
 'psycopg2-binary>=2.9.1,<3.0.0',
 'requests-cache>=0.9.3,<0.10.0',
 'requests>=2.26.0,<3.0.0',
 'scipy>=1.7.1,<2.0.0',
 'tqdm>=4.62.1,<5.0.0']

entry_points = \
{'console_scripts': ['bolster = bolster.cli:main']}

setup_kwargs = {
    'name': 'bolster',
    'version': '0.1.2',
    'description': "Bolster's Brain, you've been warned",
    'long_description': '=======\nBolster\n=======\n\n\n.. image:: https://img.shields.io/pypi/v/bolster.svg\n    :target: https://pypi.python.org/pypi/bolster\n\n.. image:: https://travis-ci.com/andrewbolster/bolster.svg?branch=main\n    :target: https://travis-ci.com/andrewbolster/bolster\n\n.. image:: https://readthedocs.org/projects/bolster/badge/?version=latest\n    :target: https://bolster.readthedocs.io/en/latest/?version=latest\n    :alt: Documentation Status\n\n.. image:: https://pyup.io/repos/github/andrewbolster/bolster/shield.svg\n    :target: https://pyup.io/repos/github/andrewbolster/bolster/\n    :alt: Updates\n\n.. image:: https://requires.io/github/andrewbolster/bolster/requirements.svg?branch=main\n    :target: https://requires.io/github/andrewbolster/bolster/requirements/?branch=main\n    :alt: Requirements Status\n\nBolster\'s Brain, you\'ve been warned\n\n* Free software: GNU General Public License v3\n* Documentation: https://bolster.readthedocs.io.\n\n\nFeatures\n--------\n\n* Efficient tree/node traversal and iteration\n* Datetime helpers\n* Concurrecy Helpers\n* Web safe Encapsulation/Decapsulation helpers\n* `pandas`-esque `aggregate`/`transform_r` functions\n* "Best Practice" AWS service handling\n\nData Sources\n------------\n* UK Companies House Listings\n* NI House Price Index Data Wrangling\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `audreyr/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`audreyr/cookiecutter-pypackage`: https://github.com/audreyr/cookiecutter-pypackage\n',
    'author': 'Andrew Bolster',
    'author_email': 'me@andrewbolster.info',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/andrewbolster/bolster',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<3.10',
}


setup(**setup_kwargs)



import io
import zipfile

import requests


def download_extract_zip(url):
    """
    Download a ZIP file and extract its contents in memory
    yields (filename, file-like object) pairs
    """
    with requests.get(url, stream=True) as response:
        response.raise_for_status()
        with zipfile.ZipFile(io.BytesIO(response.content)) as thezip:
            for zipinfo in thezip.infolist():
                with thezip.open(zipinfo) as thefile:
                    yield zipinfo.filename, thefile



from typing import AnyStr
from typing import Dict

import pandas as pd
from pandas._typing import FilePathOrBuffer


def save_xls(dict_df: Dict[AnyStr, pd.DataFrame], path: FilePathOrBuffer):
    """
    Save a dictionary of dataframes to an excel file, with each dataframe as a separate page
    """

    with pd.ExcelWriter(path) as writer:
        for key in dict_df:
            dict_df[key].to_excel(writer, key)
